using Microsoft.AspNetCore.Authentication.JwtBearer;
using Microsoft.AspNetCore.Authorization;
using Microsoft.IdentityModel.Tokens;
using System.IdentityModel.Tokens.Jwt;
using System.Security.Claims;
using System.Text;

namespace TestApi {

	//This user class can hold various required user data,
	//Such as Username, password, etc...
	public class User {
		public string Username { get; set; }
	}

	public class Program {

		//Used to generate JWT
		private static JwtSecurityTokenHandler JwtTokenGenerator = new JwtSecurityTokenHandler();

		//The JWT issuer, probably the host domain name.
		//Normally retrieved from a configuration file.
		private const string Issuer = "TestApi";

		//The api's private encryption key.
		//This should obviously not be embedded here.
		private static SymmetricSecurityKey Key = new SymmetricSecurityKey
			(Encoding.UTF8.GetBytes("97D173BBB957AC71A4C21AD12EEC797D173BBB957AC71A4C21AD12EEC797D173BBB957AC71A4C21AD12EEC797D173BBB957AC71A4C21AD12EEC7"));

		public static void Main(string[] args) {

			var builder = WebApplication.CreateBuilder(args);

			//Setup the authentication service
			builder.Services.AddAuthentication(o => {
				//Sets the default authentication scheme to Jwt
				o.DefaultAuthenticateScheme = JwtBearerDefaults.AuthenticationScheme;
				o.DefaultChallengeScheme = JwtBearerDefaults.AuthenticationScheme;
				o.DefaultScheme = JwtBearerDefaults.AuthenticationScheme;
			}).AddJwtBearer(o => {
				o.TokenValidationParameters = new TokenValidationParameters {
					IssuerSigningKey = Key,
					ValidateIssuerSigningKey = true,

					ValidIssuer = Issuer,
					ValidateIssuer = true,

					ValidateAudience = false,
					
					RequireExpirationTime = true,
					ValidateLifetime = true,
					ClockSkew = TimeSpan.Zero
				};
			});

			//Setup the authorization Service
			builder.Services.AddAuthorization(o => { 
				o.AddPolicy("SpecialAccess", policy => 
					policy.RequireClaim("AccessLevel", "Special"));
			});

			var app = builder.Build();

			app.UseRouting();
			app.UseAuthentication();
			app.UseAuthorization();

			//unrestricted access to all endpoint
			app.MapGet("/unrestricted", [AllowAnonymous] () => { 
				return "Hello world!";
			});

			//restricted access to anyone who meets the "SpecialAccess" policy
			app.MapPost("/restricted", (HttpContext httpContext) => {
				return "Restricted Content!";
			}).RequireAuthorization("SpecialAccess");

			//jwt generation endpoint
			app.MapPost("/auth", [AllowAnonymous](User user) => {

				//user defined criteria that fail authentication
				//could be username-password authentication
				if(user.Username == "AuthFail") return Results.Unauthorized();

				//Setting up parameters for a new jwt
				var tokenDescriptor = new SecurityTokenDescriptor {
					Subject = new ClaimsIdentity(new[]{
						new Claim(JwtRegisteredClaimNames.Sub, user.Username),
						
						//Grant special access to the user with username "SpecialUser"
						new Claim("AccessLevel", user.Username == "SpecialUser" ? "Special" : "Common")
					}),

					Expires = DateTime.UtcNow.AddMinutes(5),
					Issuer = Issuer,
					SigningCredentials = new SigningCredentials
						(Key, SecurityAlgorithms.HmacSha512Signature)
				};

				//Generate jwt
				SecurityToken token = JwtTokenGenerator.CreateToken(tokenDescriptor);
				string jwt = JwtTokenGenerator.WriteToken(token);
				return Results.Ok(jwt);
				
			});

			app.Run();
		}
	}
}